//-----------------------------------------------------------------------------
// name: bk-gfx.cpp
// desc: graphics stuff for bokeh visualization
//
// author: Ge Wang (ge@ccrma.stanford.edu)
//   date: 2013
//-----------------------------------------------------------------------------
#include "ff_gfx.h"
#include "ff_globals.h"
#include "ff_sim.h"
#include "ff_audio.h"
#include "melody.h"
#include "theme.h"
#include "variation.h"

#include "x-fun.h"
#include "x-gfx.h"
#include "x-loadlum.h"
#include "x-vector3d.h"
//#include "../Cube.h"

#include <iostream>
#include <vector>
using namespace std;



// bokehs
vector<YBokeh *> g_bokehs;


//-----------------------------------------------------------------------------
// function prototypes
//-----------------------------------------------------------------------------
void idleFunc();
void displayFunc();
void reshapeFunc( int width, int height );
void keyboardFunc( unsigned char, int, int );
void mouseFunc( int button, int state, int x, int y );
void specialFunc( int key, int x, int y );

void initialize_graphics();
void initialize_simulation();
bool initialize_data();
void loadTextures();
bool checkTexDim( int dim );
void setupTexture( const char * filename, GLuint texobj,
                  GLenum minFilter, GLenum maxFilter,
                  int useMipMaps = 0 );

void renderBackground();
void blendPane();
void updateNodeEntities();
void renderNodeEntities();




//-----------------------------------------------------------------------------
// name: ff_gfx_init( )
// desc: graphics/interaction entry point
//-----------------------------------------------------------------------------
bool ff_gfx_init( int argc, const char ** argv )
{
#ifdef __APPLE__
    // save working dir
    char * cwd = getcwd( NULL, 0 );
    // set path
    // #ifdef __VQ_RELEASE__
#if 1
    // HACK: use this hard-coded path
    Globals::path = "./";
#else
    // store this globally
    Globals::path = cwd;
#endif
    // compute the datapath
    Globals::datapath = Globals::path + Globals::relpath;
#endif

    // initialize GLUT
    glutInit( &argc, (char **)argv );

#ifdef __APPLE__
    //restore working dir
    chdir( cwd );
    free( cwd );
#endif

    // print about
    ff_about();
    ff_endline();

    // double buffer, use rgb color, enable depth buffer
    glutInitDisplayMode( GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH );
    // initialize the window size
    glutInitWindowSize( Globals::windowWidth, Globals::windowHeight );
    // set the window postion
    glutInitWindowPosition( 100, 100 );
    // create the window
    glutCreateWindow( "FantasiticFantasy");
    // full screen
    if( Globals::fullscreen )
        glutFullScreen();

    // set the idle function - called when idle
    glutIdleFunc( idleFunc );
    // set the display function - called when redrawing
    glutDisplayFunc( displayFunc );
    // set the reshape function - called when client area changes
    glutReshapeFunc( reshapeFunc );
    // set the keyboard function - called on keyboard events
    glutKeyboardFunc( keyboardFunc );
    // set the mouse function - called on mouse stuff
    glutMouseFunc( mouseFunc );
    // for arrow keys, etc
	glutSpecialFunc (specialFunc );

    // do our own initialization
    initialize_graphics();
    // simulation
    initialize_simulation();
    // do data
    if( !initialize_data() )
    {
        // done
        return false;
    }

    // print keys
    ff_endline();
    ff_keys();
    ff_line();
    ff_endline();

    return true;
}




//-----------------------------------------------------------------------------
// name: ff_gfx_loop( )
// desc: hand off to graphics loop
//-----------------------------------------------------------------------------
void ff_gfx_loop()
{
    // let GLUT handle the current thread from here
    glutMainLoop();
}




//-----------------------------------------------------------------------------
// Name: initialize_graphics( )
// Desc: sets initial OpenGL states and initializes any application data
//-----------------------------------------------------------------------------
void initialize_graphics()
{
    // log
    cerr << "[Fantastic Fantasy]: initializing graphics system..." << endl;

    // reset time
    XGfx::resetCurrentTime();
    // set simulation speed
    XGfx::setDeltaFactor( 1.0f );
    // get the first
    XGfx::getCurrentTime( true );
    // random
    XFun::srand();

    // set the GL clear color - use when the color buffer is cleared
    glClearColor( 0, 0, 1, 1);
    // set the shading model to 'smooth'
    glShadeModel( GL_SMOOTH );
    // enable depth
    glEnable( GL_DEPTH_TEST );
    // set the front faces of polygons
//    glFrontFace( GL_CCW );
    // set fill mode
//    glPolygonMode( GL_FRONT_AND_BACK, Globals::fillmode );
    // enable lighting
    glEnable( GL_LIGHTING );
    // enable lighting for front
    glLightModeli( GL_FRONT_AND_BACK, GL_TRUE );

    // enable color
    glEnable( GL_COLOR_MATERIAL );


    // enable transparency
    glEnable (GL_BLEND);
    glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

    // enable light 0
    glEnable( GL_LIGHT0 );

    // setup and enable light 1
    glLightfv( GL_LIGHT1, GL_AMBIENT, Globals::light1_ambient );
    glLightfv( GL_LIGHT1, GL_DIFFUSE, Globals::light1_diffuse );
    glLightfv( GL_LIGHT1, GL_SPECULAR, Globals::light1_specular );
    glEnable( GL_LIGHT1 );




    // bg color

    Globals::bgColor.set(Vector3D(0,.7,1));

//
    // clear the color buffer once
    glClear( GL_COLOR_BUFFER_BIT );
}




//-----------------------------------------------------------------------------
// name: initialize_simulation( )
// desc: simulation setup
//-----------------------------------------------------------------------------
void initialize_simulation()
{
    // instantiate simulation
    Globals::sim = new BKSim();


}




//-----------------------------------------------------------------------------
// name: initialize_data( )
// desc: load initial data
//-----------------------------------------------------------------------------
bool initialize_data()
{

    // done
    return true;
}




//-----------------------------------------------------------------------------
// name: ff_about()
// desc: ...
//-----------------------------------------------------------------------------
void ff_about()
{
    ff_line();
    fprintf( stderr, "FantasticFantasy\n" );
    ff_line();
    fprintf( stderr, "   | by Melanie Goldstein\n" );
    fprintf( stderr, "   | Stanford University | CCRMA\n" );
    fprintf( stderr, "   | http://ccrma.stanford.edu/n" );
    fprintf( stderr, "   | version: %s\n", Globals::version.c_str() );
}




//-----------------------------------------------------------------------------
// name: ff_keys()
// desc: ...
//-----------------------------------------------------------------------------
void ff_keys()
{
    ff_line();
    fprintf( stderr, "FantasticFantasy: run-time control\n" );
    ff_line();
    fprintf( stderr, "  'h' - print this help message\n" );
    fprintf( stderr, "  's' - toggle fullscreen\n" );
    fprintf( stderr, "  'q' - quit\n" );
    fprintf( stderr, " 'spacebar' - play/pause\n");
    fprintf( stderr, "  'b' - black screen\n" );
    fprintf( stderr, "  'z' - toggle mode\n" );
    fprintf( stderr, "  'x' - turn off beat/ metronome in edit mode\n" );
    fprintf( stderr, "  'left and right arrow keys' - move to the preceding or following variation/ theme in either mode (you must enter enough notes into a melody for the right arrow to unlock the next variation screen)\n" );
    fprintf( stderr, "  '-_' - decrease tempo\n" );
    fprintf( stderr, "  '+=' - increase tempo\n" );
    fprintf( stderr, "  'Note:' - in both modes a message will tell you when you have moved from one melody to another (it will tell you both from which melody you came and where you went)\n");
    
    



}




//-----------------------------------------------------------------------------
// name: ff_help()
// desc: ...
//-----------------------------------------------------------------------------
void ff_help()
{
    ff_endline();
    ff_keys();
    ff_endline();
}




//-----------------------------------------------------------------------------
// name: ff_usage()
// desc: ...
//-----------------------------------------------------------------------------
void ff_usage()
{
    ff_line();
    fprintf( stderr, "[FantasticFantasy!]: command line arguments\n" );
    ff_line();
    fprintf( stderr, "usage: CubeFantasy --[options] [name]\n" );
    fprintf( stderr, "   [options] = help | fullscreen" );
}




//-----------------------------------------------------------------------------
// name: ff_endline()
// desc: ...
//-----------------------------------------------------------------------------
void ff_endline()
{
    fprintf( stderr, "\n" );
}




//-----------------------------------------------------------------------------
// name: ff_line()
// desc: ...
//-----------------------------------------------------------------------------
void ff_line()
{
    fprintf( stderr, "---------------------------------------------------------\n" );
}




//-----------------------------------------------------------------------------
// Name: reshapeFunc( )
// Desc: called when window size changes
//-----------------------------------------------------------------------------
void reshapeFunc( int w, int h )
{
    // save the new window size
    Globals::windowWidth = w; Globals::windowHeight = h;
    // map the view port to the client area
    glViewport( 0, 0, w, h );
    // set the matrix mode to project
    glMatrixMode( GL_PROJECTION );
    // load the identity matrix
    glLoadIdentity( );
    // set the matrix mode to modelview
    glMatrixMode( GL_MODELVIEW );
    // load the identity matrix
    glLoadIdentity( );
}




//-----------------------------------------------------------------------------
// Name: look( )
// Desc: ...
//-----------------------------------------------------------------------------
void look( )
{
    // go
    Globals::fov.interp( XGfx::delta() );
    // set the matrix mode to project
    glMatrixMode( GL_PROJECTION );
    // load the identity matrix
    glLoadIdentity( );
    // create the viewing frustum
    gluPerspective( Globals::fov.value, (GLfloat)Globals::windowWidth / (GLfloat)Globals::windowHeight, .005, 500.0 );

    // set the matrix mode to modelview
    glMatrixMode( GL_MODELVIEW );
    // load the identity matrix
    glLoadIdentity();
    // position the view point
    gluLookAt( 0.0f,
              Globals::viewRadius.x * sin( Globals::viewEyeY.x ),
              Globals::viewRadius.x * cos( Globals::viewEyeY.x ),
              0.0f, 0.0f, 0.0f,
              0.0f, ( cos( Globals::viewEyeY.x ) < 0 ? -1.0f : 1.0f ), 0.0f );

    // set the position of the lights
    glLightfv( GL_LIGHT0, GL_POSITION, Globals::light0_pos );
    glLightfv( GL_LIGHT1, GL_POSITION, Globals::light1_pos );
}




//-----------------------------------------------------------------------------
// Name: keyboardFunc( )
// Desc: key event
//-----------------------------------------------------------------------------
void keyboardFunc( unsigned char key, int x, int y )
{
    // system keys (handled first)
    switch( key )
    {
        case 'q'://quit
        {
            exit( 0 );
            break;
        }

        case 'b'://black screen
        {
            static GLfloat blendAlpha = 0.15f;
            Globals::blendScreen = !Globals::blendScreen;
            if( Globals::blendScreen )
            {
                Globals::blendAlpha.goal = blendAlpha;
                Globals::blendAlpha.slew = .5f;
            }
            else
            {
                blendAlpha = Globals::blendAlpha.goal;
                Globals::blendAlpha.goal = 1;
            }
            fprintf( stderr, "[bokeh]: blendscreen:%s\n", Globals::blendScreen ? "ON" : "OFF" );
            break;

        }
       

        case '+': //with +=, -_ speed up or slow down the tempo
        case '=':
        {
            Globals::mainclock.scaleTempo(true);
            break;
        }
        case '-':
        case '_':
        {
            Globals::mainclock.scaleTempo(false);
            break;
        }
            

        case 'z': //change mode
        {
            
            if (Globals::mode == FF_EDIT) {
                Globals::mode = FF_PLAYBACK;
                 Globals::bgColor.set(Vector3D(0,0,.78));


            } else {
                Globals::mode = FF_EDIT;
                Globals::bgColor.set(Vector3D(0,.7,1));

            }
            break;
        }

        case 'x': //turn on/ off metronome
        {
            if (Globals::modes == BEAT) {
                Globals::modes = NOBEAT;
                
                
            } else {
                Globals::modes = BEAT;
                
            }
            break;
        }
        case 'h':
        {
            ff_help();
            break;
        }
      
        case 's': //toggle fullscreen
        {
            if( !Globals::fullscreen )
            {
                Globals::lastWindowWidth = Globals::windowWidth;
                Globals::lastWindowHeight = Globals::windowHeight;
                glutFullScreen();
            }
            else
                glutReshapeWindow( Globals::lastWindowWidth, Globals::lastWindowHeight );

            Globals::fullscreen = !Globals::fullscreen;
            fprintf( stderr, "[bokeh]: fullscreen:%s\n", Globals::fullscreen ? "ON" : "OFF" );
            break;
        }
       
        case ' ':
            if (!Globals::mainclock.isPlaying) {
                Globals::mainclock.play();
            } else {
                Globals::mainclock.pause();
            }

            break;

    }




    // do a reshape since viewEyeY might have changed
    reshapeFunc( Globals::windowWidth, Globals::windowHeight );
    // post redisplay
    glutPostRedisplay( );
}


//-----------------------------------------------------------------------------
// Name: processHits( )
// Desc: does the actual picking
//-----------------------------------------------------------------------------
void processHits (GLint hits, GLuint buffer[])
{
    GLuint * buff = (GLuint *)buffer;
    if (*buff != 2){
        cerr << "don't know how to handle " << *buff << "selections" << endl;
        return;
    }

    GLuint track = *(buff+3);
    GLuint step = *(buff+4);

    // cout << "track " << track << ", step " << step << endl;

    Globals::bank.getCurrentMelody()->toggleNote(track, step);
}


//-----------------------------------------------------------------------------
// Name: mouseFunc( )
// Desc: handles mouse stuff
//-----------------------------------------------------------------------------
void mouseFunc( int button, int state, int x, int y )
{

   
    if (state == GLUT_UP) {
        // 1. enter OpenGL selection mode

        GLint viewport[4];

        glSelectBuffer(PICK_BUFSIZE, Globals::selectBuf);
        glRenderMode(GL_SELECT);

        glMatrixMode(GL_PROJECTION);
        glPushMatrix();
        glLoadIdentity();

        glGetIntegerv(GL_VIEWPORT,viewport);
        gluPickMatrix(x,viewport[3]-y, 10, 10, viewport);

        // NOTE: this should always match the gluPerspective call in the look() method
        // gluPerspective( Globals::fov.value,
        //                 (GLfloat)Globals::windowWidth / (GLfloat)Globals::windowHeight, .005,
        //                 500.0 );
        glMatrixMode(GL_MODELVIEW);
        // glLoadIdentity();
        glInitNames();


        // 2. render (it will render to the selection)
        int hits;
        Globals::bank.render();


        // 3. leave OpenGL selection mode

        // restoring the original projection matrix
        glMatrixMode(GL_PROJECTION);
        glPopMatrix();
        glMatrixMode(GL_MODELVIEW);
        glFlush();

        // returning to normal rendering mode
        hits = glRenderMode(GL_RENDER);

        // if there are hits process them
        // cerr << "hits: " << hits << endl;
        if (hits != 0)
            processHits(hits, Globals::selectBuf);

    }


    glutPostRedisplay( );

}




//-----------------------------------------------------------------------------
// Name: specialFunc( )
// Desc: handles arrow stuff
//-----------------------------------------------------------------------------
void specialFunc( int key, int x, int y )
{
    // check
    bool handled = false;

    // if not handled
    if( !handled )
    {
        switch( key ) //moves from one melody ot the next in both modes if users have created long enough melodies (with enough notes)
        {
            case GLUT_KEY_RIGHT:
                if (Globals::mode==FF_EDIT) {
                if (Globals::bank.getCurrentMelody()->isComplete()) {
                    cerr<<"moved from melody " << Globals::bank.getCurrentMelody()->melody_id <<endl;

                    Globals::bank.active_melody = ++Globals::bank.active_melody % Globals::bank.melodic_input.size();

                    cerr << "moved to melody " << Globals::bank.getCurrentMelody()->melody_id << endl;

                    Globals::bank.getCurrentMelody()->makeSkeleton();
                   
                } else {
                    cerr<< "you have not completed the melody." <<endl;
                }
                } else {
                    
                        cerr<<"moved from melody " << Globals::bank.getCurrentMelody()->melody_id <<endl;
                        
                        Globals::bank.active_melody = ++Globals::bank.active_melody % Globals::bank.melodic_input.size();
                        
                        cerr << "moved to melody " << Globals::bank.getCurrentMelody()->melody_id << endl;
                        
                        Globals::bank.getCurrentMelody()->makeSkeleton();
                    }
                
                break;
    
            case GLUT_KEY_LEFT:
                Globals::bank.active_melody = --Globals::bank.active_melody % Globals::bank.melodic_input.size();
                break;
            case GLUT_KEY_UP:
                break;
            case GLUT_KEY_DOWN:
                break;
        }
    }
}




//-----------------------------------------------------------------------------
// Name: idleFunc( )
// Desc: callback from GLUT
//-----------------------------------------------------------------------------
void idleFunc( )
{
    // render the scene
    glutPostRedisplay( );
}




//-----------------------------------------------------------------------------
// Name: displayFunc( )
// Desc: callback function invoked to draw the client area
//-----------------------------------------------------------------------------
void displayFunc( )
{
    // get current time (once per frame)
    XGfx::getCurrentTime( true );



    // clear or blend
    if( Globals::blendScreen && Globals::blendAlpha.value > .0001 )
    {
        // clear the depth buffer
        glClear( GL_DEPTH_BUFFER_BIT );
        // blend screen
        blendPane();
    }
    else
    {
        // set the GL clear color - use when the color buffer is cleared
        glClearColor( Globals::bgColor.actual().x, Globals::bgColor.actual().y, Globals::bgColor.actual().z, 1.0f );
        // clear the color and depth buffers
        glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
    }

    // enable depth test
    glEnable( GL_DEPTH_TEST );

    // save state
    glPushMatrix();

//    glScalef((float)Globals::windowHeight / Globals::windowWidth, 1, 1);
    
    Globals::bank.render();


    glPopMatrix();



    glutSwapBuffers();
}




//-----------------------------------------------------------------------------
// name: blendPane()
// desc: blends a pane into the current scene
//-----------------------------------------------------------------------------
void blendPane()
{
    // enable blending
    glEnable( GL_BLEND );
    glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );
    // disable lighting
    glDisable( GL_LIGHTING );
    // disable depth test
    glDisable( GL_DEPTH_TEST );


    GLfloat h = 10;
    GLfloat d = -1;

    // draw the polyg
    glBegin( GL_QUADS );
    glVertex3f( -h, -h, d );
    glVertex3f( h, -h, d );
    glVertex3f( h, h, d );
    glVertex3f( -h, h, d );
    glEnd();

}




//-----------------------------------------------------------------------------
// name: renderBackground()
// desc: ...
//-----------------------------------------------------------------------------
void renderBackground()
{
    // save the current matrix
    glPushMatrix( );

    // restore
    glPopMatrix( );
}




//-------------------------------------------------------------------------------
// name: loadTexture()
// desc: load textures
//-------------------------------------------------------------------------------
void loadTextures()
{
    char filename[256];
    GLenum minFilter, maxFilter;
    int i;

    // log
    fprintf( stderr, "[bokeh]: loading textures...\n" );

    // set store alignment
    glPixelStorei( GL_UNPACK_ALIGNMENT, 1 );

    // set filter types
    minFilter = GL_LINEAR;
    maxFilter = GL_LINEAR;

    // load tng flares
    for( i = ff_TEX_FLARE_TNG_1; i <= ff_TEX_FLARE_TNG_5; i++ )
    {
        glGenTextures( 1, &(Globals::textures[i]) );
        sprintf( filename, "%sflare-tng-%d.bw", Globals::datapath.c_str(), i - ff_TEX_FLARE_TNG_1 + 1 );
        setupTexture( filename, Globals::textures[i], minFilter, maxFilter );
    }
}








//--------------------------------------------------------------------------------
// name: setupTexture()
// desc: ...
//--------------------------------------------------------------------------------
void setupTexture( const char * filename, GLuint texobj,
                  GLenum minFilter, GLenum maxFilter, int useMipMaps )
{
    unsigned char * buf = NULL;
    int width = 0, height = 0, components = 0;

    glBindTexture( GL_TEXTURE_2D, texobj );
    glTexEnvf( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE );
    glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, minFilter );
    glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, maxFilter );

    // log
    // fprintf( stderr, "[bokeh]: loading %s...\n", filename );

    // load luminance
    buf = loadLuminance( filename, &width, &height, &components );

    // log
    // fprintf( stderr, "[bokeh]: '%s' : %dx%dx%d\n", filename, width, height, components);

    // build mip maps
    if( useMipMaps )
    {
        gluBuild2DMipmaps( GL_TEXTURE_2D, 1, width, height,
                          GL_LUMINANCE, GL_UNSIGNED_BYTE, buf );
    }
    else
    {
        glTexImage2D( GL_TEXTURE_2D, 0, 1, width, height, 0,
                     GL_LUMINANCE, GL_UNSIGNED_BYTE, buf );
    }

    free(buf);
}




//-----------------------------------------------------------------------------
// name: checkTexDim( )
// desc: checks to see if a dim is a valid opengl texture dimension
//-----------------------------------------------------------------------------
bool checkTexDim( int dim )
{
    if( dim < 0 )
        return false;

    int i, count = 0;

    // count bits
    for( i = 0; i < 31; i++ )
        if( dim & ( 0x1 << i ) )
            count++;

    // this is true only if dim is power of 2
    return count == 1;
}
